/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// le problème des N reines

#include "prelude.h"
/* #include "sort.h" */

// la ligne k est-elle compatible avec les lignes précédentes ?
bool check(int n, int sol[], int k) {
  for (int i = 0; i < k; i++)
    if (sol[i] == sol[k] || abs(sol[i] - sol[k]) == abs(i - k))
      return false;
  return true;
}

// retour sur trace

// entrée : 0 <= k <= n et une solution partielle dans sol[0..k[
// sortie : true, si sol a pu être totalement complétée
//          false si ce n'est pas possible, et sol[0..k[ est inchangée
bool solve(int n, int sol[], int k) {
  if (k == n) return true;
  for (int v = 0; v < n; v++) {
    sol[k] = v;
    if (check(n, sol, k) && solve(n, sol, k+1))
      return true;
  }
  return false;
}

void print(int n, int sol[]) {
  for (int i = 0; i < n; i++) {
    printf("|");
    for (int j = 0; j < n; j++) {
      printf("%s", (sol[i] == j) ? "Q|" : ".|");
    }
    printf("\n");
  }
}

// compter toutes les solutions
int count(int n, int sol[], int k) {
  if (k == n) return 1;
  int s = 0;
  for (int v = 0; v < n; v++) {
    sol[k] = v;
    if (check(n, sol, k))
      s += count(n, sol, k+1);
  }
  return s;
}

void queens(int n) {
  printf("n = %d\n", n);
  int *sol = calloc(n, sizeof(int));
  if (solve(n, sol, 0)) {
    print(n, sol);
    printf("  %d solutions\n", count(n, sol, 0));
  } else
    printf("  pas de solution !\n");
}

// version probabiliste

// entrée : 0 <= k <= n et une solution partielle dans sol[0..k[
// sortie : true si sol a pu être complétée avec une solution
//          false sinon (mais ne veut pas dire qu'il n'y a pas de solution)

// note : on fait un tirage équiprobable parmi toutes les colonnes
// encore possibles pour la ligne k en un seul parcours
bool solve_prob(int n, int sol[], int k) {
  if (k == n) return true;
  int c = 0; // colonne candidate
  int t = 0; // nombre de candidats
  for (int v = 0; v < n; v++) {
    sol[k] = v;
    if (check(n, sol, k)) {
      t++;
      if (rand () % t == 0)
        c = v;
    }
  }
  if (t == 0) return false;
  sol[k] = c;
  return solve_prob(n, sol, k+1);
}

int attempts = 0;

void queens_prob(int n, int sol[]) {
  attempts = 0;
  while (true) {
    attempts++;
    if (solve_prob(n, sol, 0))
      return;
  }
}

int main(int argc, char **argv) {
  /* for (int n = 1; n < 10; n++) */
  /*   queens(n); */
  int n = 30;
  if (argc > 1) n = atoi(argv[1]);
  int *sol = calloc(n, sizeof(int));
  queens_prob(n, sol);
  print(n, sol);
  printf("%d tentatives\n", attempts);
}

